#ifndef __ABWXMLTOKENS_HXX__
#define __ABWXMLTOKENS_HXX__

const int XML_A = 1;
const int XML_ABIWORD = 2;
const int XML_AWML = 3;
const int XML_BOOKMARK = 4;
const int XML_BR = 5;
const int XML_C = 6;
const int XML_CBR = 7;
const int XML_CELL = 8;
const int XML_D = 9;
const int XML_DATA = 10;
const int XML_ENDNOTE = 11;
const int XML_F = 12;
const int XML_FIELD = 13;
const int XML_FOOT = 14;
const int XML_FRAME = 15;
const int XML_HISTORY = 16;
const int XML_I = 17;
const int XML_IGNOREDWORDS = 18;
const int XML_IMAGE = 19;
const int XML_IW = 20;
const int XML_L = 21;
const int XML_LISTS = 22;
const int XML_M = 23;
const int XML_METADATA = 24;
const int XML_P = 25;
const int XML_PAGESIZE = 26;
const int XML_PBR = 27;
const int XML_REVISIONS = 28;
const int XML_S = 29;
const int XML_SECTION = 30;
const int XML_STYLES = 31;
const int XML_TABLE = 32;
const int XML_VERSION = 33;

const int XML_TOKEN_COUNT = 33;

const int XML_TOKEN_INVALID = -1;

#endif
